#include "fifteen_objects.h"
#include "fifteen_resources.h"

//--------------------- User code ---------------------//
const unsigned short TOP_OFFSET = 55;
const unsigned int TIMER_RESET_VALUE = 5535; //0x0BDB;
const unsigned short TRUE = 1;
const unsigned short FALSE = 0;
unsigned short HalfASecond = 0;
unsigned short UpdateTime = 0;

typedef struct GameState{
  unsigned short GameMatrix[4][4];
  unsigned int TimeElapsed;
  unsigned short GameStarted;
  unsigned short GamePrepared;
}TGameState;

TGameState newGameState;

void DisplayTimeElapsed(){
  unsigned short minutes, seconds;
  unsigned char txt[5] = "0000";
  unsigned short i;

  minutes = newGameState.TimeElapsed / 60;
  seconds = newGameState.TimeElapsed % 60;
  ShortToStr(minutes, txt);
  Label6_Caption[0] = txt[2];
  Label6_Caption[1] = txt[3];

  ShortToStr(seconds, txt);
  Label6_Caption[2] = ':';
  Label6_Caption[3] = txt[2];
  Label6_Caption[4] = txt[3];
  Label6_Caption[5] = 0;
  for (i = 0; i < 5; i++){
    if (Label6_Caption[i] == ' ')
      Label6_Caption[i] = '0';
  }
  strcpy(Label51_Caption, Label6_Caption);
  TFT_Set_Pen(StatusBar.Color, 0);
  TFT_Set_Brush(1, StatusBar.Color, 0, 0, 0, 0);
  TFT_Rectangle(Label6.Left, Label6.Top, Label6.Left+Label6.Width, Label6.Top+Label6.Height);
  DrawLabel(&Label6);
}

static void ResetTimer(unsigned short DrawIt){
  newGameState.TimeElapsed = 0;
  if (DrawIt)
    DisplayTimeElapsed();

  // Timer Interrupt every 0.5s;
  HalfASecond = 0;
  //TMR1 = TIMER_RESET_VALUE;
  TMR0L = 0b10011111; //Lo(TIMER_RESET_VALUE);
  TMR0H = 0b00010101; //Hi(TIMER_RESET_VALUE);
  //T1CON = 0b0010000000110000;
  T0CON = 0b00000000; //Timer off, 16-bit, internal clock source, prescaler assigned, 1:2 prescaler
  //T1IF_bit = 0;
  T0IF_bit = 0;
  //T1IE_bit = 0;
  T0IE_bit = 0;
}

static void ContinueTimer(){
  //TON_T1CON_bit = 1;
  TMR0ON_bit = 1;
  //T1IF_bit = 0;
  T0IF_bit = 0;
  //T1IE_bit = 1;
  T0IE_bit = 1;
  
  GIE_bit = 1;
}

static void PauseTimer(){
  //TON_T1CON_bit = 0;
  TMR0ON_bit = 0;
  //T1IE_bit = 0;
  T0IE_bit = 0;
  //T1IF_bit = 0;
  T0IF_bit = 0;
}

static void StopTimer(unsigned short DrawIt){
  newGameState.TimeElapsed = 0;
  //T1IE_bit = 0;
  T0IE_bit = 0;
  //T1IF_bit = 0;
  T0IF_bit = 1;
  //T1CON = 0;
  TMR0ON_bit = 0;
  if (DrawIt)
    DisplayTimeElapsed();
}

static void TimerInterrupt() iv 0x000008 { //IVT_ADDR_T1INTERRUPT{
  //TMR1 = TIMER_RESET_VALUE;
  TMR0L = 0b10011111; //Lo(TIMER_RESET_VALUE);
  TMR0H = 0b00010101; //Hi(TIMER_RESET_VALUE);
  if (T0IF_bit == 1) {
    if (newGameState.GameStarted){
      if (HalfASecond == 100){ //48,000,000 Hz / 4 (System clock) = 12,000,000 / 2 prescaler = 6,000,000 Hz / 60,000 TMR0 counts to overflow = 100 Hz (100 times per second)
        HalfASecond = 0;
        newGameState.TimeElapsed++;
        //DisplayTimeElapsed();
        UpdateTime = 1;
      }
      else
        //HalfASecond = 1;
        HalfASecond++;
    }
    //T1IF_bit = 0;
    T0IF_bit = 0;
  }
}

static void DrawGame(){
  unsigned short i, j;
  unsigned short CurrentButton;
  unsigned short temp;
  //temp.B0 = T1IE_bit;
  temp.B0 = T0IE_bit;
  //T1IE_bit = 0;
  T0IE_bit = 0;
  //T1IF_bit = 0;
  T0IF_bit = 0;
  for (i = 0; i < 4; i++){
    for (j = 0; j < 4; j++){
      CurrentButton = newGameState.GameMatrix[i][j];
      if (CurrentButton != 0){
        MainScreen.Buttons[CurrentButton-1]->Top = i*60 + TOP_OFFSET;
        MainScreen.Buttons[CurrentButton-1]->Left = j*60;
        DrawButton(MainScreen.Buttons[CurrentButton-1]);
      }
      else{
        Empty.Top = i*60 + TOP_OFFSET;
        Empty.Left = j*60;
        DrawButton(&Empty);
      }
    }
  }
  //T1IE_bit = temp.B0;
  T0IE_bit = temp.B0;
}

void InitGame(){
  newGameState.GameStarted = 0;
  newGameState.GamePrepared = 0;
  ResetTimer(FALSE);
}

void ShowVictoryScreen(){
  StopTimer(FALSE);
  DrawScreen(&Victory);
}

static void IsGameOver(){
  unsigned short i, j, counter;
  unsigned short Victory;

  Victory = TRUE;
  counter = 1;

  for (i = 0; i < 4; i++){
    for (j = 0; j < 4; j++){
      if (newGameState.GameMatrix[i][j] == counter){
        counter++;
        if (counter == 16)
          counter = 0;
      }
      else {
        Victory = FALSE;
        break;
      }
    }
  }

  if (Victory){
    newGameState.GamePrepared = 0;
    newGameState.GameStarted = 0;
    ShowVictoryScreen();
  }
}

static void CalculateMove(unsigned short ButtonNumber, TButton* Button){
  short ButtonXpos, ButtonYPos;
  short EmptyXpos, EmptyYPos;
  unsigned short CanMove = 0;
  unsigned short Temp;

  if (newGameState.GamePrepared){
    if (!newGameState.GameStarted){
      ResetTimer(FALSE);
      ContinueTimer();
      newGameState.GameStarted = 1;
    }

    ButtonXpos = Button->Left / Button->Width;
    ButtonYPos = (Button->Top - TOP_OFFSET) / Button->Height;

    EmptyXpos = Empty.Left / Empty.Width;
    EmptyYPos = (Empty.Top - TOP_OFFSET) / Empty.Height;

    if ((ButtonXpos - 1 == EmptyXpos) && (ButtonYPos == EmptyYPos))
     CanMove = 1;
    else if ((ButtonXpos + 1 == EmptyXpos) && (ButtonYPos == EmptyYPos))
     CanMove = 1;
    else if ((ButtonXpos == EmptyXpos) && (ButtonYPos - 1 == EmptyYPos))
     CanMove = 1;
    else if ((ButtonXpos == EmptyXpos) && (ButtonYPos + 1 == EmptyYPos))
     CanMove = 1;
    else
      CanMove = 0;

    if (CanMove){
      newGameState.GameMatrix[ButtonYpos][ButtonXpos] = 0;
      newGameState.GameMatrix[EmptyYpos][EmptyXpos] = ButtonNumber;
      Temp = Empty.Left;
      Empty.Left =  Button->Left;
      Button->Left = Temp;

      Temp = Empty.Top;
      Empty.Top = Button->Top;
      Button->Top = Temp;
      PauseTimer();
      DrawButton(&Empty);
      DrawButton(Button);
      IsGameOver();
      ContinueTimer();
    }
  }
}
//----------------- End of User code ------------------//

// Event Handlers

void Button1Down() {
  CalculateMove(1, &Button1);
}

void Button2Down() {
  CalculateMove(2, &Button2);
}

void Button3Down() {
  CalculateMove(3, &Button3);
}

void Button4Down() {
  CalculateMove(4, &Button4);
}

void Button5Down() {
  CalculateMove(5, &Button5);
}

void Button6Down() {
  CalculateMove(6, &Button6);
}

void Button7Down() {
  CalculateMove(7, &Button7);
}

void Button8Down() {
  CalculateMove(8, &Button8);
}

void Button9Down() {
  CalculateMove(9, &Button9);
}

void Button10Down() {
  CalculateMove(10, &Button10);
}

void Button11Down() {
  CalculateMove(11, &Button11);
}

void Button12Down() {
  CalculateMove(12, &Button12);
}

void Button13Down() {
  CalculateMove(13, &Button13);
}

void Button14Down() {
  CalculateMove(14, &Button14);
}

void Button15Down() {
  CalculateMove(15, &Button15);
}

void Randomise(){
  unsigned short i, j;
  unsigned short temp;

  srand(rand());

  for (i = 0; i < 4; i++)
    for (j = 0; j < 4; j++)
      newGameState.GameMatrix[i][j] = 0;

  for (i = 1; i < 16; i++){
    temp = rand() % 15;
    while (newGameState.GameMatrix[temp / 4][temp % 4] != 0)
      temp = rand() % 15;
    newGameState.GameMatrix[temp / 4][temp % 4] = i;
  }
}

unsigned short CheckParity() {
  unsigned short numbers[16];
  unsigned short i, j, cnt, parity;
  unsigned short temp;

  cnt = 0;
  for (i = 0; i < 4; i++) {
    for (j = 0; j < 4; j++) {
      numbers[cnt] = newGameState.GameMatrix[i][j];
      cnt++;
    }
  }

  cnt = 0;
  parity = 0;
  while (cnt < 14) {
    if (numbers[cnt] > numbers[cnt + 1]) {
      temp = numbers[cnt];
      numbers[cnt] = numbers[cnt + 1];
      numbers[cnt + 1] = temp;
      cnt = 0;
      parity++;
    } else {cnt++;}
  }

  return parity % 2;
}

void NewGame(){
  InitGame();
  do {Randomise();} while (CheckParity() != 0);
  //Randomise();
  DrawGame();
  DisplayTimeElapsed();
  newGameState.GamePrepared = 1;
}

void ShowHelpScreen(){
  PauseTimer();
  DrawScreen(&Help);
}

void ShowMainScreen(){
  DrawScreen(&MainScreen);
  if (newGameState.GameStarted)
    ContinueTimer();
}